# -*- coding: utf-8 -*-

####################################################################
# FlexDB
# Guenther Starnberger <gst@sysfrog.org>
####################################################################

# Python / Zope ####################################################
import string
import zipfile
import csv
from StringIO import StringIO
from AccessControl import ClassSecurityInfo
####################################################################

# Plone / Archetypes ###############################################
from Products.Archetypes.public import OrderedBaseFolderSchema, Schema
from Products.Archetypes.public import OrderedBaseFolder, registerType
from Products.ATSchemaEditorNG.SchemaEditor import SchemaEditor
from Products.Archetypes.public import TextField, TextAreaWidget
from Products.Archetypes.utils import OrderedDict
from Products.PageTemplates.PageTemplate import PageTemplate
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
####################################################################

# FlexDB ###########################################################
from FlexDBItem import FlexDBItem
from FlexDBCategory import FlexDBCategory
from Products.FlexDB.FancyImageField import FancyImageField
import config
####################################################################

class FlexDB(SchemaEditor, OrderedBaseFolder):
    '''This is the FlexDB which is able to contain FlexDBItems'''

    __implements__ = (SchemaEditor.__implements__, OrderedBaseFolder.__implements__)

    meta_type = portal_type = 'FlexDB'

    security = ClassSecurityInfo()
    
    # Factory Type Information
    filter_content_types = True # Only allow specific sub-items
    allowed_content_types = ('FlexDBItem', 'FlexDBCategory', ) # this sub-items are allowed

    immediate_view = 'flexdb_list'
    default_view   = 'flexdb_list'
    content_icon   = 'flexdb_icon.gif'

    # Automatically generate a short name
    _at_rename_after_creation = True

    # The FlexDB class doesn't need a specific schema so we directly
    # copy the base-schema instead
    schema = OrderedBaseFolderSchema.copy()

    actions = ( {'id': 'view',
                 'name': 'View',
                 'action': 'string:${object_url}/flexdb_list',
                 'permissions': (config.view_permission,)},
                {'id': 'edit_schema',
                 'name': 'Schema Editor',
                 'action': 'flexdb_editor',
                 'permissions': (config.edit_permission,)},
                {'id': 'flexdb_maintenance',
                 'name': 'Maintenance',
                 'action': 'flexdb_maintenance',
                 'permissions': (config.edit_permission,)},)

    def __init__(self, oid, **kwargs):
        OrderedBaseFolder.__init__(self, oid, **kwargs)

    security.declarePrivate('manage_afterAdd')
    def manage_afterAdd(self, item, container):
        '''post creation (or post renaming) actions'''
        
        OrderedBaseFolder.manage_afterAdd(self, item, container)

        self.atse_init()
        self.atse_registerObject( FlexDBItem,
                                  filtered_schemas = ('metadata', 'Main'),
                                  undeleteable_fields = config.UNDELETEABLE_FIELDS,
                                  domain = 'FlexDB' )

        # ATSENG currently doesn't allow to use a different
        # configuration data in products. Therefore we have to
        # manually copy the settings from our configuration file after
        # the ATSENG initialization.

        self._field_registry = config.FIELD_INFO
        
        self._widget_registry = OrderedDict()
        for k, v in config.WIDGET_REGISTRY_INFO:
                self._widget_registry[k] = v

        self._field_registry = OrderedDict()
        for k, v in config.FIELD_REGISTRY_INFO:
                self._field_registry[k] = v

        self._storage_registry = OrderedDict()
        for k, v in config.STORAGE_REGISTRY_INFO:
                self._storage_registry[k] = v

    security.declareProtected(config.ManageSchemaPermission, 'atse_update')
    def atse_update(self, schema_id, schema_template, fielddata, REQUEST, RESPONSE=None):
        '''TTW Editor for Archetyp Schema'''

        # We use a simplified Webinterface - so we have to map our
        # type to the more complicated ATSENG type.
        try:
            for attribute in ['type', 'widget', 'storage']:
                setattr(fielddata, attribute, config.FLEXDB_TYPES[fielddata['flexdbtype']][attribute])
        except KeyError:
            # When creating a new field we get a KeyError exception which we can ignore
            pass

        result = SchemaEditor.atse_update(self, schema_id, schema_template, fielddata, REQUEST, RESPONSE)

        # Only call block when the arguments contain fielddata.name
        # (when no new field is created). Basically the same reason as
        # with the exception in the block above.
        if fielddata.has_key('name'):
            field = self._schemas['FlexDBItem'][fielddata.name]

            if not fielddata.has_key('visible_list'):
                field.widget.visible['list'] = 'invisible'
            else:
                field.widget.visible['list'] = 'visible'

            # Make sure that changes are persistent
            self._p_changed = True

        return result

    security.declareProtected(config.ManageSchemaPermission, 'atse_updateManagedSchemas')
    def atse_updateManagedSchema(self, portal_type, schema_template, REQUEST=None, RESPONSE=None):
        '''Also reindex objects when Schema is updated'''

        # The original ATSENG code (currently) doesn't reindex the
        # objects when the schema is modified. But this is required -
        # when we remove e.g. a textfield we don't want to have the
        # content of this field in the index. Reindexing helps.

        result = SchemaEditor.atse_updateManagedSchema(self, portal_type, schema_template, REQUEST, RESPONSE)

        [object.getObject().reindexObject() for object in self.portal_catalog.searchResults(portal_type = portal_type)]

        return result

    security.declareProtected(config.ManageSchemaPermission, 'get_flexdb_type')
    def get_flexdb_type(self, field):
        '''get current flexdb type'''

        # Extracts type, widget and attribute information from a field
        # and returns the corresponding FlexDB type

        type_name = self.atse_getFieldType(field)
        widget_name = field.getWidgetName()
        storage_name = field.getStorageName(field)

        flexdb_type = self.rev_types(config.FLEXDB_TYPES)[(type_name, widget_name, storage_name)]

        return flexdb_type

    security.declareProtected(config.ManageSchemaPermission, 'get_all_flexdb_types')
    def get_all_flexdb_types(self):
        '''returns all FlexDB types'''

        return config.FLEXDB_TYPES.keys()

    security.declareProtected(config.view_permission, 'search_items')
    def search_items(self, context, term = ''):
        '''filter result list'''

        # Search or list the contained FlexDB items. If no term is
        # specified it is ignored. This needs to be called from a Page
        # Template as we need the context in which this method is
        # called. Searching is only done in sub-folders of the context
        # - this is required for the category support.

        brains = self.portal_catalog.searchResults(Type = 'Flex DBItem',
                                                   review_state = 'published',
                                                   path = string.join(context.getPhysicalPath(), '/'),
                                                   SearchableText = term)
        return [brain.getObject() for brain in brains]

    security.declareProtected(config.view_permission, 'get_schema_fields')
    def get_schema_fields(self, mode = None):
        '''get fields of schema'''

        # Return all visible fields of a schema

        return [field for field in
                self._schemas['FlexDBItem'].filterFields(isMetadata=0)
                if field.widget.isVisible(None, mode) == 'visible']

    security.declarePrivate('rev_types')
    def rev_types(self, types):
        '''Reverse the given type mappings'''

        # We only store the mapping of FlexDB -> ATSENG types in one
        # direction. This method calculates the mappings in the other
        # direction.

        rev = {}
        
        for key in types.keys():
            value = types[key]
            rev[(value['type'], value['widget'] + 'Widget', value['storage'] + 'Storage')] = key
            
        return rev

    security.declareProtected(config.edit_permission, 'csv_export')
    def csv_export(self, plaintext = False, REQUEST = None):
        '''Export the database to a CSV file'''

        # Exports a CSV file of the contained data. The id of items in
        # sub-categories is exported as relative path -
        # e.g. category1/itemid. If the 'plaintext' attribute is set
        # the content type is left on the Plone default. Otherwise
        # text/comma-separated-values is used.

        self_path = self.getPhysicalPath()

        csv_buffer = StringIO()
        writer = csv.writer(csv_buffer)
        
        field_list = self.get_schema_fields()
        id_field = field_list.pop(0)

        for item in self.search_items(self):
            row = []

            absolute_item_path = item.getPhysicalPath()
            relative_item_path = absolute_item_path[len(self_path):]
            
            row.append(string.join(relative_item_path, '/'))
            
            for field in field_list:
                if hasattr(field, 'flexdb_export'):
                    data = field.flexdb_export(item)
                else:
                    data = field.getRaw(item)
                row.append(data)
                
            writer.writerow(row)

        result = csv_buffer.getvalue()
        csv_buffer.close()

        if REQUEST is not None and plaintext == False:
            REQUEST.RESPONSE.setHeader('Content-Type', 'text/comma-separated-values; charset=UTF-8')
            REQUEST.RESPONSE.setHeader('Content-Disposition', 'attachment; filename="export.csv"')

        return result
    
    security.declareProtected(config.edit_permission, 'csv_import')
    def csv_import(self, file, publish = None):
        '''Import from a CSV file'''

        # Imports a CSV file (which complies to the FlexDB CSV
        # format). Optionally automatically publishes it. The id of
        # items in sub-categories needs to be specified as relative
        # path - e.g. category1/itemid.

        csv_buffer = StringIO(file.read())
        reader = csv.reader(csv_buffer)

        field_list = self.get_schema_fields()

        work_flag = False

        for row in reader:
            categories = row.pop(0).split('/')
            id = categories.pop()

            context = self.get_or_create_category(self, categories, publish)
            
            context.invokeFactory('FlexDBItem', id)
            obj = context._getOb(id)

            for field in field_list[1:]:
                data = row.pop(0)
                if hasattr(field, 'flexdb_import'):
                    field.flexdb_import(obj, data)
                else:
                    field.set(obj, data)

            if publish is not None:
                self.portal_workflow.doActionFor(obj, 'publish', 'plone_workflow')

            obj.reindexObject()

            if not work_flag:
                work_flag = True

        csv_buffer.close()

        if work_flag:
            msg = 'Items successfully imported'
        else:
            msg = 'No Items were found'

        return self.flexdb_maintenance(msg = msg)

    security.declarePrivate('get_or_create_category')
    def get_or_create_category(self, context, category_path, publish = None):
        '''Helper function for csv_import'''

        # This method recursively searches for a category and creates
        # it if it doesn't exist.

        for category in category_path:
            if not hasattr(context, category):
                context.invokeFactory('FlexDBCategory', category)
                category_obj = context._getOb(category)
                category_obj.setTitle(category)
                
                if publish is not None:
                    self.portal_workflow.doActionFor(category_obj, 'publish', 'plone_workflow')

                category_obj.reindexObject()

                context = category_obj
            else:
                context = getattr(context, category)

        return context

    security.declarePrivate('get_relative_path')
    def get_relative_path(self, context, object):

        absolute_item_path = object.getPhysicalPath()
        self_path = context.getPhysicalPath()
        relative_item_path = absolute_item_path[len(self_path):]
        relative_item_path_string = string.join(relative_item_path, '/')

        return relative_item_path

    security.declareProtected(config.view_permission, 'get_relative_path_string')
    def get_relative_path_string(self, context, object):

        relative_item_path = self.get_relative_path(context, object)
        relative_item_path_string = string.join(relative_item_path, '/')

        return relative_item_path_string

    security.declarePrivate('get_relative_path_to_string')
    def relative_path_to_string(self, path):

        result = string.join(path, '/')

        if result.startswith('/'):
            result = result[1:]

        return result

    security.declareProtected(config.edit_permission, 'webui_export')
    def webui_export(self, current_context, original_context, zip):
        'export in webui format'

        # to avoid redundancy this method is also called by
        # FlexDBCategory using Acquisition, therefore current_context
        # must be used instead of self.

        self_relative_path = self.get_relative_path(original_context, current_context)
        self_zip_path = self.relative_path_to_string(self_relative_path + ('index.html',))
        zip.writestr(self_zip_path, str(self.flexdb_webui_main(context = current_context, original_context = original_context)))

        for item in current_context.listFolderContents():
            item.webui_export(item, original_context, zip)
            
        return None

    security.declareProtected(config.edit_permission, 'html_export')
    def html_export(self, REQUEST = None):
        'Export content as HTML'

        zipdata = StringIO()
        zip = zipfile.ZipFile(zipdata, 'w', zipfile.ZIP_DEFLATED)
        self.webui_export(self, self, zip)
        zip.close()
        result = zipdata.getvalue()
        zipdata.close()

        if REQUEST is not None:
            REQUEST.RESPONSE.setHeader('Content-Type', 'application/zip')
            REQUEST.RESPONSE.setHeader('Content-Disposition', 'attachment; filename="webui.zip"')

        return result

    security.declareProtected(config.edit_permission, 'image_regenerate_thumnails')
    def image_regenerate_thumnails(self):
        'Regenerate Thumbnails'

        field_list = self.get_schema_fields()

        for item in self.search_items(self):
            for field in field_list:
                if type(field) == FancyImageField:
                    field.removeScales(item)
                    field.createScales(item)

        return self.flexdb_maintenance(msg = 'Thumbnails regenerated')

    security.declarePrivate('flexdb_lookup_provider')
    def flexdb_lookup_provider(self):
        'We are the Schema provider'
        return self

registerType(FlexDB)
